@@ -42,17 +42,17 @@ Created 11/5/1995 Heikki Tuuri
 #include "srv0start.h"
 #include "srv0srv.h"
 #include "log.h"
-#include "mariadb_stats.h"
 
 TRANSACTIONAL_TARGET
 bool buf_pool_t::page_hash_contains(const page_id_t page_id, hash_chain &chain)
+	noexcept
 {
   transactional_shared_lock_guard<page_hash_latch> g
     {page_hash.lock_get(chain)};
   return page_hash.get(page_id, chain);
 }
 
-/** If there are buf_pool.curr_size per the number below pending reads, then
+/** If there are buf_pool.curr_size() per the number below pending reads, then
 read-ahead is not done: this is to prevent flooding the buffer pool with
 i/o-fixed buffer blocks */
 #define BUF_READ_AHEAD_PEND_LIMIT	2
@@ -71,22 +71,12 @@ and the lock released later.
 @param block      preallocated buffer block (set to nullptr if consumed)
 @return pointer to the block
 @retval	nullptr in case of an error */
-TRANSACTIONAL_TARGET
 static buf_page_t *buf_page_init_for_read(const page_id_t page_id,
                                           ulint zip_size,
                                           buf_pool_t::hash_chain &chain,
-                                          buf_block_t *&block)
+                                          buf_block_t *&block) noexcept
 {
-  buf_page_t *bpage= nullptr;
-  if (!zip_size || (zip_size & 1))
-  {
-    bpage= &block->page;
-    block->initialise(page_id, zip_size & ~1, buf_page_t::READ_FIX);
-    /* x_unlock() will be invoked
-    in buf_page_t::read_complete() by the io-handler thread. */
-    block->page.lock.x_lock(true);
-  }
-
+  buf_page_t *bpage= !zip_size || (zip_size & 1) ? &block->page : nullptr;
   page_hash_latch &hash_lock= buf_pool.page_hash.lock_get(chain);
   hash_lock.lock();
   if (buf_pool.page_hash.get(page_id, chain))
@@ -94,13 +84,6 @@ static buf_page_t *buf_page_init_for_read(const page_id_t page_id,
 page_exists:
     hash_lock.unlock();
     /* The page is already in the buffer pool. */
-    if (bpage)
-    {
-      bpage->lock.x_unlock(true);
-      ut_d(mysql_mutex_lock(&buf_pool.mutex));
-      ut_d(bpage->set_state(buf_page_t::MEMORY));
-      ut_d(mysql_mutex_unlock(&buf_pool.mutex));
-    }
     return nullptr;
   }
 
@@ -121,6 +104,11 @@ page_exists:
   if (UNIV_LIKELY(bpage != nullptr))
   {
     block= nullptr;
+    reinterpret_cast<buf_block_t*>(bpage)->
+      initialise(page_id, zip_size & ~1, buf_page_t::READ_FIX);
+    /* x_unlock() will be invoked
+    in buf_page_t::read_complete() by the io-handler thread. */
+    bpage->lock.x_lock(true);
     /* Insert into the hash table of file pages */
     buf_pool.page_hash.append(chain, bpage);
     hash_lock.unlock();
@@ -173,15 +161,17 @@ page_exists:
     page_zip_set_size(&bpage->zip, zip_size);
     bpage->zip.data = (page_zip_t*) data;
 
+    /* Because bpage is a compressed-only block descriptor, it cannot be
+    passed to buf_pool.page_guess(), and therefore there is no risk of a
+    a false match. Therefore, we can safely initialize bpage before
+    acquiring hash_lock. */
     bpage->lock.init();
     bpage->init(buf_page_t::READ_FIX, page_id);
     bpage->lock.x_lock(true);
 
-    {
-      transactional_lock_guard<page_hash_latch> g
-        {buf_pool.page_hash.lock_get(chain)};
-      buf_pool.page_hash.append(chain, bpage);
-    }
+    hash_lock.lock();
+    buf_pool.page_hash.append(chain, bpage);
+    hash_lock.unlock();
 
     /* The block must be put to the LRU list, to the old blocks.
     The zip size is already set into the page zip */
@@ -196,18 +186,28 @@ func_exit:
   return bpage;
 }
 
+inline ulonglong mariadb_measure() noexcept
+{
+#if (MY_TIMER_ROUTINE_CYCLES)
+  return my_timer_cycles();
+#else
+  return my_timer_microseconds();
+#endif
+}
+
 /** Low-level function which reads a page asynchronously from a file to the
 buffer buf_pool if it is not already there, in which case does nothing.
 Sets the io_fix flag and sets an exclusive lock on the buffer frame. The
 flag is cleared and the x-lock released by an i/o-handler thread.
 
 @param[in] page_id	page id
-@param[in] zip_size	ROW_FORMAT=COMPRESSED page size, or 0,
-			bitwise-ORed with 1 in recovery
+@param[in] zip_size	0 or ROW_FORMAT=COMPRESSED page size
+			bitwise-ORed with 1 to allocate an uncompressed frame
 @param[in,out] chain	buf_pool.page_hash cell for page_id
 @param[in,out] space	tablespace
 @param[in,out] block	preallocated buffer block
-@param[in] sync		true if synchronous aio is desired
+@param[in] thd		current_thd if sync
+@param[in] sync		whether synchronous aio is desired
 @return error code
 @retval DB_SUCCESS if the page was read
 @retval DB_SUCCESS_LOCKED_REC if the page exists in the buffer pool already */
@@ -219,7 +219,8 @@ buf_read_page_low(
 	buf_pool_t::hash_chain&	chain,
 	fil_space_t*		space,
 	buf_block_t*&		block,
-	bool			sync = false)
+	THD*			thd = nullptr,
+	bool			sync = false) noexcept
 {
 	buf_page_t*	bpage;
 
@@ -237,14 +238,12 @@ buf_read_page_low(
 
 	ut_ad(bpage->in_file());
 	ulonglong mariadb_timer = 0;
+	trx_t *const trx= thd ? thd_to_trx(thd) : nullptr;
 
-	if (sync) {
-		thd_wait_begin(nullptr, THD_WAIT_DISKIO);
-		if (const ha_handler_stats *stats = mariadb_stats) {
-			if (stats->active) {
-				mariadb_timer = mariadb_measure();
-			}
-		}
+	thd_wait_begin(thd, THD_WAIT_DISKIO);
+
+	if (trx && trx->active_handler_stats) {
+		mariadb_timer = mariadb_measure();
 	}
 
 	DBUG_LOG("ib_buf",
@@ -261,17 +260,16 @@ buf_read_page_low(
 			     dst, bpage);
 
 	if (UNIV_UNLIKELY(fio.err != DB_SUCCESS)) {
+		recv_sys.free_corrupted_page(page_id, *space->chain.start);
 		buf_pool.corrupted_evict(bpage, buf_page_t::READ_FIX);
 	} else if (sync) {
-		thd_wait_end(nullptr);
+		thd_wait_end(thd);
 		/* The i/o was already completed in space->io() */
 		fio.err = bpage->read_complete(*fio.node);
 		space->release();
-		if (fio.err == DB_FAIL) {
-			fio.err = DB_PAGE_CORRUPTED;
-		}
 		if (mariadb_timer) {
-			mariadb_increment_pages_read_time(mariadb_timer);
+			trx->active_handler_stats->pages_read_time
+				+= mariadb_measure() - mariadb_timer;
 		}
 	}
 
@@ -295,6 +293,24 @@ static void buf_read_release(buf_block_t *block)
   }
 }
 
+/** Report a completed read-ahead batch.
+@param space  tablespace
+@param count  number of pages submitted for reading */
+static ATTRIBUTE_NOINLINE
+void buf_read_ahead_report(const fil_space_t &space, size_t count) noexcept
+{
+  if (THD *thd= current_thd)
+    if (trx_t *trx= thd_to_trx(thd))
+      if (ha_handler_stats *stats= trx->active_handler_stats)
+        stats->pages_prefetched+= count;
+  mysql_mutex_lock(&buf_pool.mutex);
+  /* Read ahead is considered one I/O operation for the purpose of
+     LRU policy decision. */
+  buf_LRU_stat_inc_io();
+  buf_pool.stat.n_ra_pages_read_rnd+= count;
+  mysql_mutex_unlock(&buf_pool.mutex);
+}
+
 /** Applies a random read-ahead in buf_pool if there are at least a threshold
 value of accessed pages from the random read-ahead area. Does not read any
 page, not even the one at the position (space, offset), if the read-ahead
@@ -303,10 +319,9 @@ pages: to avoid deadlocks this function must be written such that it cannot
 end up waiting for these latches!
 @param[in]	page_id		page id of a page which the current thread
 wants to access
-@param[in]	zip_size	ROW_FORMAT=COMPRESSED page size, or 0
 @return number of page read requests issued */
 TRANSACTIONAL_TARGET
-ulint buf_read_ahead_random(const page_id_t page_id, ulint zip_size)
+ulint buf_read_ahead_random(const page_id_t page_id) noexcept
 {
   if (!srv_random_read_ahead || page_id.space() >= SRV_TMP_SPACE_ID)
     /* Disable the read-ahead for temporary tablespace */
@@ -317,7 +332,7 @@ ulint buf_read_ahead_random(const page_id_t page_id, ulint zip_size)
     return 0;
 
   if (os_aio_pending_reads_approx() >
-      buf_pool.curr_size / BUF_READ_AHEAD_PEND_LIMIT)
+      buf_pool.curr_size() / BUF_READ_AHEAD_PEND_LIMIT)
     return 0;
 
   fil_space_t* space= fil_space_t::get(page_id.space());
@@ -353,6 +368,7 @@ read_ahead:
 
   /* Read all the suitable blocks within the area */
   buf_block_t *block= nullptr;
+  unsigned zip_size{space->zip_size()};
   if (UNIV_LIKELY(!zip_size))
   {
   allocate_block:
@@ -382,51 +398,32 @@ read_ahead:
   }
 
   if (count)
-  {
-    mariadb_increment_pages_prefetched(count);
-    DBUG_PRINT("ib_buf", ("random read-ahead %zu pages from %s: %u",
-			  count, space->chain.start->name,
-			  low.page_no()));
-    mysql_mutex_lock(&buf_pool.mutex);
-    /* Read ahead is considered one I/O operation for the purpose of
-    LRU policy decision. */
-    buf_LRU_stat_inc_io();
-    buf_pool.stat.n_ra_pages_read_rnd+= count;
-    mysql_mutex_unlock(&buf_pool.mutex);
-  }
+    buf_read_ahead_report(*space, count);
 
   space->release();
   buf_read_release(block);
   return count;
 }
 
-/** High-level function which reads a page from a file to buf_pool
-if it is not already there. Sets the io_fix and an exclusive lock
-on the buffer frame. The flag is cleared and the x-lock
-released by the i/o-handler thread.
-@param page_id    page id
-@param zip_size   ROW_FORMAT=COMPRESSED page size, or 0
-@param chain      buf_pool.page_hash cell for page_id
-@retval DB_SUCCESS if the page was read and is not corrupted,
-@retval DB_SUCCESS_LOCKED_REC if the page was not read
-@retval DB_PAGE_CORRUPTED if page based on checksum check is corrupted,
-@retval DB_DECRYPTION_FAILED if page post encryption checksum matches but
-after decryption normal page checksum does not match.
-@retval DB_TABLESPACE_DELETED if tablespace .ibd file is missing */
-dberr_t buf_read_page(const page_id_t page_id, ulint zip_size,
-                      buf_pool_t::hash_chain &chain)
+dberr_t buf_read_page(const page_id_t page_id,
+                      buf_pool_t::hash_chain &chain, bool unzip) noexcept
 {
   fil_space_t *space= fil_space_t::get(page_id.space());
-  if (!space)
+  if (UNIV_UNLIKELY(!space))
   {
-    ib::info() << "trying to read page " << page_id
-               << " in nonexisting or being-dropped tablespace";
+    sql_print_information("InnoDB: trying to read page "
+                          "[page id: space=" UINT32PF
+                          ", page number=" UINT32PF "]"
+                          " in nonexisting or being-dropped tablespace",
+                          page_id.space(), page_id.page_no());
     return DB_TABLESPACE_DELETED;
   }
 
   /* Our caller should already have ensured that the page does not
   exist in buf_pool.page_hash. */
   buf_block_t *block= nullptr;
+  unsigned zip_size= space->zip_size();
+
   if (UNIV_LIKELY(!zip_size))
   {
   allocate_block:
@@ -435,13 +432,14 @@ dberr_t buf_read_page(const page_id_t page_id, ulint zip_size,
     block= buf_LRU_get_free_block(have_mutex);
     mysql_mutex_unlock(&buf_pool.mutex);
   }
-  else if (recv_recovery_is_on())
+  else if (unzip)
   {
     zip_size|= 1;
     goto allocate_block;
   }
 
-  dberr_t err= buf_read_page_low(page_id, zip_size, chain, space, block, true);
+  dberr_t err= buf_read_page_low(page_id, zip_size, chain, space, block,
+                                 current_thd, true);
   buf_read_release(block);
   return err;
 }
@@ -454,7 +452,7 @@ released by the i/o-handler thread.
 @param[in]	page_id		page id
 @param[in]	zip_size	ROW_FORMAT=COMPRESSED page size, or 0 */
 void buf_read_page_background(fil_space_t *space, const page_id_t page_id,
-                              ulint zip_size)
+                              ulint zip_size) noexcept
 {
   buf_pool_t::hash_chain &chain= buf_pool.page_hash.cell_get(page_id.fold());
   if (buf_pool.page_hash_contains(page_id, chain))
@@ -511,10 +509,8 @@ NOTE 2: the calling thread may own latches on pages: to avoid deadlocks this
 function must be written such that it cannot end up waiting for these
 latches!
 @param[in]	page_id		page id; see NOTE 3 above
-@param[in]	zip_size	ROW_FORMAT=COMPRESSED page size, or 0
 @return number of page read requests issued */
-TRANSACTIONAL_TARGET
-ulint buf_read_ahead_linear(const page_id_t page_id, ulint zip_size)
+ulint buf_read_ahead_linear(const page_id_t page_id) noexcept
 {
   /* check if readahead is disabled.
   Disable the read ahead logic for temporary tablespace */
@@ -526,7 +522,7 @@ ulint buf_read_ahead_linear(const page_id_t page_id, ulint zip_size)
     return 0;
 
   if (os_aio_pending_reads_approx() >
-      buf_pool.curr_size / BUF_READ_AHEAD_PEND_LIMIT)
+      buf_pool.curr_size() / BUF_READ_AHEAD_PEND_LIMIT)
     return 0;
 
   const uint32_t buf_read_ahead_area= buf_pool.read_ahead_area;
@@ -553,6 +549,11 @@ fail:
     return 0;
   }
 
+  if (trx_sys_hdr_page(page_id))
+    /* If it is an ibuf bitmap page or trx sys hdr, we do no
+    read-ahead, as that could break the ibuf page access order */
+    goto fail;
+
   /* How many out of order accessed pages can we ignore
   when working out the access pattern for linear readahead */
   ulint count= std::min<ulint>(buf_pool_t::READ_AHEAD_PAGES -
@@ -647,6 +648,7 @@ failed:
 
   /* If we got this far, read-ahead can be sensible: do it */
   buf_block_t *block= nullptr;
+  unsigned zip_size{space->zip_size()};
   if (UNIV_LIKELY(!zip_size))
   {
   allocate_block:
@@ -678,18 +680,7 @@ failed:
   }
 
   if (count)
-  {
-    mariadb_increment_pages_prefetched(count);
-    DBUG_PRINT("ib_buf", ("random read-ahead %zu pages from %s: %u",
-                          count, space->chain.start->name,
-                          new_low.page_no()));
-    mysql_mutex_lock(&buf_pool.mutex);
-    /* Read ahead is considered one I/O operation for the purpose of
-    LRU policy decision. */
-    buf_LRU_stat_inc_io();
-    buf_pool.stat.n_ra_pages_read+= count;
-    mysql_mutex_unlock(&buf_pool.mutex);
-  }
+    buf_read_ahead_report(*space, count);
 
   space->release();
   buf_read_release(block);
@@ -702,7 +693,7 @@ failed:
 @param recs     log records
 @param init_lsn page initialization, or 0 if the page needs to be read */
 void buf_read_recover(fil_space_t *space, const page_id_t page_id,
-                      page_recv_t &recs, lsn_t init_lsn)
+                      page_recv_t &recs, lsn_t init_lsn) noexcept
 {
   ut_ad(space->id == page_id.space());
   space->reacquire();
