@@ -24,6 +24,7 @@ Purge old versions
 Created 3/26/1996 Heikki Tuuri
 *******************************************************/
 
+#define MYSQL_SERVER
 #include "trx0purge.h"
 #include "fsp0fsp.h"
 #include "mach0data.h"
@@ -42,6 +43,7 @@ Created 3/26/1996 Heikki Tuuri
 #include <mysql/service_thd_mdl.h>
 #include <mysql/service_wsrep.h>
 #include "log.h"
+#include "sql_class.h"
 
 /** Maximum allowable purge history length.  <=0 means 'infinite'. */
 ulong		srv_max_purge_lag = 0;
@@ -135,28 +137,11 @@ void purge_sys_t::close()
 /** Determine if the history of a transaction is purgeable.
 @param trx_id  transaction identifier
 @return whether the history is purgeable */
-TRANSACTIONAL_TARGET bool purge_sys_t::is_purgeable(trx_id_t trx_id) const
+bool purge_sys_t::is_purgeable(trx_id_t trx_id) const noexcept
 {
-  bool purgeable;
-#if !defined SUX_LOCK_GENERIC && !defined NO_ELISION
-  purgeable= false;
-  if (xbegin())
-  {
-    if (!latch.is_write_locked())
-    {
-      purgeable= view.changes_visible(trx_id);
-      xend();
-    }
-    else
-      xabort();
-  }
-  else
-#endif
-  {
-    latch.rd_lock(SRW_LOCK_CALL);
-    purgeable= view.changes_visible(trx_id);
-    latch.rd_unlock();
-  }
+  latch.rd_lock(SRW_LOCK_CALL);
+  bool purgeable= view.changes_visible(trx_id);
+  latch.rd_unlock();
   return purgeable;
 }
 
@@ -353,7 +338,7 @@ void purge_sys_t::rseg_enable(trx_rseg_t &rseg)
 inline dberr_t purge_sys_t::iterator::free_history_rseg(trx_rseg_t &rseg) const
 {
   fil_addr_t hdr_addr;
-  mtr_t mtr;
+  mtr_t mtr{nullptr};
   bool freed= false;
   uint32_t rseg_ref= 0;
   const auto last_boffset= srv_page_size - TRX_UNDO_LOG_OLD_HDR_SIZE;
@@ -629,7 +614,7 @@ function is called, the caller
 (purge_coordinator_callback or purge_truncation_callback)
 must not have any latches on undo log pages!
 */
-TRANSACTIONAL_TARGET void trx_purge_truncate_history()
+void trx_purge_truncate_history()
 {
   ut_ad(purge_sys.head <= purge_sys.tail);
   purge_sys_t::iterator &head= purge_sys.head.trx_no
@@ -716,7 +701,7 @@ not_free:
 
     log_free_check();
 
-    mtr_t mtr;
+    mtr_t mtr{nullptr};
     mtr.start();
     mtr.x_lock_space(space);
     /* Associate the undo tablespace with mtr.
@@ -779,7 +764,7 @@ not_free:
   }
 }
 
-buf_block_t *purge_sys_t::get_page(page_id_t id)
+buf_block_t *purge_sys_t::get_page(page_id_t id, trx_t *trx)
 {
   ut_ad(!recv_sys.recovery_on);
 
@@ -788,7 +773,7 @@ buf_block_t *purge_sys_t::get_page(page_id_t id)
 
   if (!undo_page)
   {
-    undo_page= buf_pool.page_fix(id); // batch_cleanup() will unfix()
+    undo_page= buf_pool.page_fix(id, trx); // batch_cleanup() will unfix()
     if (!undo_page)
       pages.erase(id);
     else
@@ -798,7 +783,7 @@ buf_block_t *purge_sys_t::get_page(page_id_t id)
   return undo_page;
 }
 
-bool purge_sys_t::rseg_get_next_history_log()
+bool purge_sys_t::rseg_get_next_history_log(trx_t *trx) noexcept
 {
   fil_addr_t prev_log_addr;
 
@@ -810,7 +795,7 @@ bool purge_sys_t::rseg_get_next_history_log()
   next_stored= false;
 
   if (buf_block_t *undo_page=
-      get_page(page_id_t(rseg->space->id, rseg->last_page_no)))
+      get_page(page_id_t(rseg->space->id, rseg->last_page_no), trx))
   {
     const byte *log_hdr= undo_page->page.frame + rseg->last_offset();
     prev_log_addr= flst_get_prev_addr(log_hdr + TRX_UNDO_HISTORY_NODE);
@@ -831,7 +816,7 @@ bool purge_sys_t::rseg_get_next_history_log()
     /* Read the previous log header. */
     trx_id_t trx_no= 0;
     if (const buf_block_t* undo_page=
-        get_page(page_id_t(rseg->space->id, prev_log_addr.page)))
+        get_page(page_id_t(rseg->space->id, prev_log_addr.page), trx))
     {
       const byte *log_hdr= undo_page->page.frame + prev_log_addr.boffset;
       trx_no= mach_read_from_8(log_hdr + TRX_UNDO_TRX_NO);
@@ -854,13 +839,14 @@ bool purge_sys_t::rseg_get_next_history_log()
   }
 
   rseg->latch.wr_unlock();
-  return choose_next_log();
+  return choose_next_log(trx);
 }
 
 /** Position the purge sys "iterator" on the undo record to use for purging.
+@param trx  transaction attached to current_thd
 @retval false when nothing is to be purged
 @retval true  when purge_sys.rseg->latch was locked */
-bool purge_sys_t::choose_next_log()
+bool purge_sys_t::choose_next_log(trx_t *trx) noexcept
 {
   ut_ad(!next_stored);
 
@@ -904,7 +890,7 @@ bool purge_sys_t::choose_next_log()
   else
   {
     page_id_t id{rseg->space->id, hdr_page_no};
-    buf_block_t *b= get_page(id);
+    buf_block_t *b= get_page(id, trx);
     if (!b)
       goto purge_nothing;
     const trx_undo_rec_t *undo_rec=
@@ -919,7 +905,7 @@ bool purge_sys_t::choose_next_log()
       if (next == FIL_NULL)
         goto purge_nothing;
       id.set_page_no(next);
-      b= get_page(id);
+      b= get_page(id, trx);
       if (!b)
         goto purge_nothing;
       undo_rec=
@@ -940,11 +926,14 @@ bool purge_sys_t::choose_next_log()
 
 /**
 Get the next record to purge and update the info in the purge system.
+@param trx                transaction attached to current_thd
 @param roll_ptr           undo log pointer to the record
 @return buffer-fixed reference to undo log record
 @retval {nullptr,1} if the whole undo log can skipped in purge
 @retval {nullptr,0} if nothing is left, or on corruption */
-inline trx_purge_rec_t purge_sys_t::get_next_rec(roll_ptr_t roll_ptr)
+inline
+trx_purge_rec_t purge_sys_t::get_next_rec(trx_t *trx, roll_ptr_t roll_ptr)
+  noexcept
 {
   ut_ad(next_stored);
   ut_ad(tail.trx_no < low_limit_no());
@@ -954,7 +943,7 @@ inline trx_purge_rec_t purge_sys_t::get_next_rec(roll_ptr_t roll_ptr)
   {
     /* It is the dummy undo log record, which means that there is no need to
     purge this undo log. Look for the next undo log and record to purge */
-    if (rseg_get_next_history_log())
+    if (rseg_get_next_history_log(trx))
       rseg->latch.wr_unlock();
     return {nullptr, 1};
   }
@@ -963,7 +952,7 @@ inline trx_purge_rec_t purge_sys_t::get_next_rec(roll_ptr_t roll_ptr)
 
   page_id_t page_id{rseg->space->id, page_no};
   bool locked= true;
-  buf_block_t *b= get_page(page_id);
+  buf_block_t *b= get_page(page_id, trx);
   if (UNIV_UNLIKELY(!b))
   {
     if (locked)
@@ -989,7 +978,7 @@ inline trx_purge_rec_t purge_sys_t::get_next_rec(roll_ptr_t roll_ptr)
     if (next != FIL_NULL)
     {
       page_id.set_page_no(next);
-      if (buf_block_t *next_page= get_page(page_id))
+      if (buf_block_t *next_page= get_page(page_id, trx))
       {
         rec2= trx_undo_page_get_first_rec(next_page, hdr_page_no, hdr_offset);
         if (rec2)
@@ -1006,7 +995,7 @@ inline trx_purge_rec_t purge_sys_t::get_next_rec(roll_ptr_t roll_ptr)
   {
   got_no_rec:
     /* Look for the next undo log and record to purge */
-    locked= rseg_get_next_history_log();
+    locked= rseg_get_next_history_log(trx);
   }
 
   if (locked)
@@ -1015,13 +1004,13 @@ inline trx_purge_rec_t purge_sys_t::get_next_rec(roll_ptr_t roll_ptr)
   return {b->page.frame + uint16_t(roll_ptr), roll_ptr};
 }
 
-inline trx_purge_rec_t purge_sys_t::fetch_next_rec()
+inline trx_purge_rec_t purge_sys_t::fetch_next_rec(trx_t *trx) noexcept
 {
   roll_ptr_t roll_ptr;
 
   if (!next_stored)
   {
-    bool locked= choose_next_log();
+    bool locked= choose_next_log(trx);
     ut_ad(locked == next_stored);
     if (!locked)
       goto got_nothing;
@@ -1046,63 +1035,70 @@ inline trx_purge_rec_t purge_sys_t::fetch_next_rec()
   }
 
   /* The following will advance the purge iterator. */
-  return get_next_rec(roll_ptr);
+  return get_next_rec(trx, roll_ptr);
 }
 
 /** Close all tables that were opened in a purge batch for a worker.
 @param node   purge task context
 @param thd    purge coordinator thread handle */
-static void trx_purge_close_tables(purge_node_t *node, THD *thd)
+static void trx_purge_close_tables(purge_node_t *node, THD *thd) noexcept
 {
   for (auto &t : node->tables)
   {
-    if (!t.second.first);
-    else if (t.second.first == reinterpret_cast<dict_table_t*>(-1));
-    else
+    dict_table_t *table= t.second.first;
+    if (table != nullptr && table != reinterpret_cast<dict_table_t*>(-1))
+      table->release();
+  }
+
+  for (auto &t : node->tables)
+  {
+    dict_table_t *table= t.second.first;
+    if (table != nullptr && table != reinterpret_cast<dict_table_t*>(-1))
     {
-      dict_table_close(t.second.first, thd, t.second.second);
       t.second.first= reinterpret_cast<dict_table_t*>(-1);
+      if (t.second.second != nullptr)
+        thd->mdl_context.release_lock(t.second.second);
     }
   }
 }
 
 void purge_sys_t::wait_FTS(bool also_sys)
 {
+  std::this_thread::yield();
   for (const uint32_t mask= also_sys ? ~0U : ~PAUSED_SYS; m_FTS_paused & mask;)
     std::this_thread::sleep_for(std::chrono::milliseconds(10));
 }
 
 __attribute__((nonnull))
-/** Aqcuire a metadata lock on a table.
+/** Acquire a metadata lock on a table.
 @param table        table handle
 @param mdl_context  metadata lock acquisition context
-@param mdl          metadata lcok
+@param mdl          metadata lock
 @return table handle
 @retval nullptr if the table is not found or accessible
 @retval -1      if the purge of history must be suspended due to DDL */
 static dict_table_t *trx_purge_table_acquire(dict_table_t *table,
                                              MDL_context *mdl_context,
-                                             MDL_ticket **mdl)
+                                             MDL_ticket **mdl) noexcept
 {
   ut_ad(dict_sys.frozen_not_locked());
   *mdl= nullptr;
 
   if (!table->is_readable() || table->corrupted)
-  {
-    table->release();
     return nullptr;
-  }
 
   size_t db_len= dict_get_db_name_len(table->name.m_name);
   if (db_len == 0)
-    return table; /* InnoDB system tables are not covered by MDL */
+  {
+    /* InnoDB system tables are not covered by MDL */
+  got_table:
+    table->acquire();
+    return table;
+  }
 
   if (purge_sys.must_wait_FTS())
-  {
   must_wait:
-    table->release();
     return reinterpret_cast<dict_table_t*>(-1);
-  }
 
   char db_buf[NAME_LEN + 1];
   char tbl_buf[NAME_LEN + 1];
@@ -1110,7 +1106,7 @@ static dict_table_t *trx_purge_table_acquire(dict_table_t *table,
 
   if (!table->parse_name<true>(db_buf, tbl_buf, &db_len, &tbl_len))
     /* The name of an intermediate table starts with #sql */
-    return table;
+    goto got_table;
 
   {
     MDL_request request;
@@ -1123,37 +1119,38 @@ static dict_table_t *trx_purge_table_acquire(dict_table_t *table,
       goto must_wait;
   }
 
-  return table;
+  goto got_table;
 }
 
 /** Open a table handle for the purge of committed transaction history
 @param table_id     InnoDB table identifier
 @param mdl_context  metadata lock acquisition context
-@param mdl          metadata lcok
+@param mdl          metadata lock
 @return table handle
 @retval nullptr if the table is not found or accessible
 @retval -1      if the purge of history must be suspended due to DDL */
 static dict_table_t *trx_purge_table_open(table_id_t table_id,
                                           MDL_context *mdl_context,
-                                          MDL_ticket **mdl)
+                                          MDL_ticket **mdl) noexcept
 {
-  dict_sys.freeze(SRW_LOCK_CALL);
+  dict_table_t *table;
 
-  dict_table_t *table= dict_sys.find_table(table_id);
-
-  if (table)
-    table->acquire();
-  else
+  for (;;)
   {
+    dict_sys.freeze(SRW_LOCK_CALL);
+    table= dict_sys.find_table(table_id);
+    if (table)
+      break;
     dict_sys.unfreeze();
     dict_sys.lock(SRW_LOCK_CALL);
     table= dict_load_table_on_id(table_id, DICT_ERR_IGNORE_FK_NOKEY);
-    if (table)
-      table->acquire();
     dict_sys.unlock();
     if (!table)
       return nullptr;
-    dict_sys.freeze(SRW_LOCK_CALL);
+    /* At this point, the freshly loaded table may already have been evicted.
+    We must look it up again while holding a shared dict_sys.latch.  We keep
+    trying this until the table is found in the cache or it cannot be found
+    in the dictionary (because the table has been dropped or rebuilt). */
   }
 
   table= trx_purge_table_acquire(table, mdl_context, mdl);
@@ -1165,17 +1162,13 @@ ATTRIBUTE_COLD
 dict_table_t *purge_sys_t::close_and_reopen(table_id_t id, THD *thd,
                                             MDL_ticket **mdl)
 {
-  MDL_context *mdl_context= static_cast<MDL_context*>(thd_mdl_context(thd));
-  ut_ad(mdl_context);
+  MDL_context *mdl_context= &thd->mdl_context;
  retry:
   ut_ad(m_active);
 
   for (que_thr_t *thr= UT_LIST_GET_FIRST(purge_sys.query->thrs); thr;
        thr= UT_LIST_GET_NEXT(thrs, thr))
-  {
-    purge_node_t *node= static_cast<purge_node_t*>(thr->child);
-    trx_purge_close_tables(node, thd);
-  }
+    trx_purge_close_tables(static_cast<purge_node_t*>(thr->child), thd);
 
   m_active= false;
   wait_FTS(false);
@@ -1213,32 +1206,23 @@ dict_table_t *purge_sys_t::close_and_reopen(table_id_t id, THD *thd,
 }
 
 /** Run a purge batch.
-@param n_purge_threads	number of purge threads
-@param thd              purge coordinator thread handle
 @param n_work_items     number of work items (currently tables) to process
 @return new purge_sys.head */
-static purge_sys_t::iterator trx_purge_attach_undo_recs(THD *thd,
-                                                        ulint *n_work_items)
+static purge_sys_t::iterator
+trx_purge_attach_undo_recs(trx_t *trx, ulint *n_work_items) noexcept
 {
-  que_thr_t *thr;
+  que_thr_t *thr= nullptr;
   purge_sys_t::iterator head= purge_sys.tail;
 
   /* Fetch and parse the UNDO records. The UNDO records are added
   to a per purge node vector. */
-  thr= nullptr;
 
   std::unordered_map<table_id_t, purge_node_t *>
     table_id_map(TRX_PURGE_TABLE_BUCKETS);
   purge_sys.m_active= true;
 
-  MDL_context *const mdl_context=
-    static_cast<MDL_context*>(thd_mdl_context(thd));
-  ut_ad(mdl_context);
-
-  const size_t max_pages=
-    std::min(buf_pool.curr_size * 3 / 4, size_t{srv_purge_batch_size});
-
-  while (UNIV_LIKELY(srv_undo_sources) || !srv_fast_shutdown)
+  for (THD *const thd{trx->mysql_thd};
+       UNIV_LIKELY(srv_undo_sources) || !srv_fast_shutdown; )
   {
     /* Track the max {trx_id, undo_no} for truncating the
     UNDO logs once we have purged the records. */
@@ -1247,7 +1231,7 @@ static purge_sys_t::iterator trx_purge_attach_undo_recs(THD *thd,
       head= purge_sys.tail;
 
     /* Fetch the next record, and advance the purge_sys.tail. */
-    trx_purge_rec_t purge_rec= purge_sys.fetch_next_rec();
+    trx_purge_rec_t purge_rec= purge_sys.fetch_next_rec(trx);
 
     if (!purge_rec.undo_rec)
     {
@@ -1265,7 +1249,7 @@ static purge_sys_t::iterator trx_purge_attach_undo_recs(THD *thd,
     if (!table_node)
     {
       std::pair<dict_table_t *, MDL_ticket *> p;
-      p.first= trx_purge_table_open(table_id, mdl_context, &p.second);
+      p.first= trx_purge_table_open(table_id, &thd->mdl_context, &p.second);
       if (p.first == reinterpret_cast<dict_table_t *>(-1))
         p.first= purge_sys.close_and_reopen(table_id, thd, &p.second);
 
@@ -1287,12 +1271,12 @@ static purge_sys_t::iterator trx_purge_attach_undo_recs(THD *thd,
       ut_ad(!table_node->in_progress);
     }
 
-    if (purge_sys.n_pages_handled() >= max_pages)
+    const size_t size{purge_sys.n_pages_handled()};
+    if (size >= size_t{srv_purge_batch_size} ||
+        size >= buf_pool.usable_size() * 3 / 4)
       break;
   }
 
-  purge_sys.m_active= false;
-
 #ifdef UNIV_DEBUG
   thr= UT_LIST_GET_FIRST(purge_sys.query->thrs);
   for (ulint i= 0; thr && i < *n_work_items;
@@ -1338,9 +1322,10 @@ static void trx_purge_wait_for_workers_to_complete()
   ut_ad(srv_get_task_queue_length() == 0);
 }
 
-TRANSACTIONAL_INLINE
 void purge_sys_t::batch_cleanup(const purge_sys_t::iterator &head)
 {
+  m_active= false;
+
   /* Release the undo pages. */
   for (auto p : pages)
     p.second->unfix();
@@ -1353,25 +1338,20 @@ void purge_sys_t::batch_cleanup(const purge_sys_t::iterator &head)
 
   /* Limit the end_view similar to what trx_purge_truncate_history() does. */
   const trx_id_t trx_no= head.trx_no ? head.trx_no : tail.trx_no;
-#ifdef SUX_LOCK_GENERIC
   end_latch.wr_lock();
-#else
-  transactional_lock_guard<srw_spin_lock_low> g(end_latch);
-#endif
   this->head= head;
   end_view= view;
   end_view.clamp_low_limit_id(trx_no);
-#ifdef SUX_LOCK_GENERIC
   end_latch.wr_unlock();
-#endif
 }
 
 /**
 Run a purge batch.
+@param trx           dummy transaction associated with the purge coordinator
 @param n_tasks       number of purge tasks to submit to the queue
 @param history_size  trx_sys.history_size()
 @return number of undo log pages handled in the batch */
-TRANSACTIONAL_TARGET ulint trx_purge(ulint n_tasks, ulint history_size)
+ulint trx_purge(trx_t *trx, ulint n_tasks, ulint history_size) noexcept
 {
   ut_ad(n_tasks > 0);
 
@@ -1379,11 +1359,10 @@ TRANSACTIONAL_TARGET ulint trx_purge(ulint n_tasks, ulint history_size)
 
   ut_d(if (srv_purge_view_update_only_debug) return 0);
 
-  THD *const thd= current_thd;
-
   /* Fetch the UNDO recs that need to be purged. */
   ulint n_work= 0;
-  const purge_sys_t::iterator head= trx_purge_attach_undo_recs(thd, &n_work);
+  THD *const thd{trx->mysql_thd};
+  const purge_sys_t::iterator head= trx_purge_attach_undo_recs(trx, &n_work);
   const size_t n_pages= purge_sys.n_pages_handled();
 
   {
